;;; crypt++-fnf.el --- file not found hook for crypt++

;; Author: Lawrence R. Dodd <dodd@roebling.poly.edu>
;;	Lars Lindberg <lli@sypro.cap.se>
;; Maintainer: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Created: 1993 Aug 25
;; Version: 1.18
;; Keywords: extensions

;;; Copyright (C) 1993 Lawrence R. Dodd

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;;; Lars Lindberg <lli@sypro.cap.se> had the idea for this package.

;;; WHAT IT DOES: 
;;; 
;;; When loaded a function in this package is attached to
;;; `find-file-not-found-hooks'.  When a find-file is done on a non-existent
;;; filename the functions of find-file-not-found-hooks are run.  This package
;;; will try searching for the filename augmented with each of the compression
;;; and encryption filename extensions used in the crypt++.el package and for
;;; the filename sans these filename extensions. 
;;; 
;;; This allows you to find files whose names are listed in other places
;;; (files, text, mail messages, etc.) without the gzip extension even though
;;; you have gzip'ed it.  Say a TAGS file is created for all `.el'-files in
;;; the GNU Emacs distribution.  With this package you can gzip all those
;;; .el-files because `M-.' (find-tag) would find it via the
;;; `find-file-not-found-hooks'.  Another example would be to locate gzip'ed
;;; header files (`.h') directly from the "#include" statement in the source
;;; code.  A final example is in dired-mode.  Say you have a file called
;;; "foobar.gz".  You use `!' (dired-do-shell-command) to gunzip it.  Dired
;;; will not revert the dired buffer so the file "foobar.gz" will still be
;;; listed.  If you then try to visit this file using `f'
;;; (dired-advertised-find-file) you would normally visit the empty file
;;; "foobar.gz" but with this code it will find the proper file "foobar."

;;; INSTALLATION:
;;;
;;; To use this package, simply put it in a file called "crypt++-fnf.el" in a
;;; lisp directory known to Emacs (see `load-path'), byte-compile it, and put
;;; the line:
;;;
;;;                      (require 'crypt++-fnf)
;;;
;;; in your ~/.emacs file or in the file default.el in the ../lisp directory.


;;; Code:

;;; Get crypt++. 
(require 'crypt++)


;;;; UNREGEXP-QUOTE 

;;; Date: Thu, 26 Aug 93 13:56:48 +0200
;;; From: prlar@aom.ericsson.se (Lars Lindberg)
;;; To: dodd@roebling.poly.edu
;;; Subject: (unregexp-quote)

(defun flatten (list)
  ;; Flattens a list. Not very effective.
  (cond
   ((null list)
    list)
   ((listp (car list))
    (append (flatten (car list)) (flatten (cdr list))))
   (t
    (cons (car list) (flatten (cdr list))))))

(defun multiply-lists (string-list string-or-list)

  ;; Takes a list of strings, LIST and combines it with
  ;; STRING-OR-LIST to a new list.
  ;;
  ;; The functionality is shown easiest by examples:
  ;; 
  ;; (multiply-lists '("a" "bc" "abd") "ef") ->
  ;;  ("aef" "bcef" "abdef")
  ;;
  ;; (multiply-lists '("a" "bc" "abd") '("ef" "gh")) ->
  ;;   ("aef" "bcef" "abdef" "aegh" "bcegh" "abdegh")

  (cond
   ;; String case
   ((stringp string-or-list)
    (mapcar (function (lambda (elt)
                        (concat elt string-or-list)))
            string-list))
   ;; Null case
   ((null string-or-list)
    string-list)
   ;; List case
   ((listp string-or-list)
    (flatten (mapcar (function (lambda (elt)
                                 ;; recursive call
                                 (multiply-lists string-list elt)))
                     string-or-list)))
   (t
    (debug)
    (error "Unknown type for STRING-OR-LIST in (multiply-lists)"))))

(defun unregexp-quote (regexp &optional asterisk-max plus-max)
  "Returns a list of all strings that match REGEXP.
If the regexp contains '*' and/or '+', then of course
is not all possible strings in the returned list.
ASTERISK-MAX says how many repetions that are allowed for '*'.
PLUS-MAX is the same for '+'."
  (interactive "*sRegexp: ")
  ;;(message regexp) (sit-for 1)
  (let ((len (length regexp))
        (i 0) (result '(""))
        j char level tmp-result (new-result "")
        )
    (while (< i len)
      (setq char (string-to-char (substring regexp i))
            i (1+ i))
      (cond
       ;; Quoted with '\'
       ((char-equal char ?\\)
        (setq char (string-to-char (substring regexp i))
              i (1+ i))
        (cond
         ;; Grouping
         ((char-equal char ?\()
          (setq result (multiply-lists result new-result)
                level 1
                j i)
          (while (and (< j len) (> level 0))
            (setq char (string-to-char (substring regexp j)))
            (cond
             ((char-equal char ?\\)
              (setq j (1+ j)
                    char (string-to-char (substring regexp j)))
              (cond
               ((char-equal char ?\))
                (setq level (1- level)))
               ((char-equal char ?\()
                (setq level (1+ level))))))
            (setq j (1+ j)))
          (and (not (= level 0)) (error "Malformed regexp."))
          (setq new-result
                (unregexp-quote (substring regexp i (- j 2))
                                asterisk-max plus-max)
                i j))
         ;; alternative ("or")
         ((char-equal char ?\|)
          (setq new-result
                (flatten (nconc (list new-result)
                                (list (unregexp-quote (substring regexp i)
                                                      asterisk-max
                                                      plus-max))))
                i len))
         ;; quote of special character
         (t
          (setq result (multiply-lists result new-result)
                new-result (char-to-string char)))))
       ;; Character set
       ((char-equal char ?\[)
        (setq result (multiply-lists result new-result)
              level 1
              j i
              new-result '())
        (while (and (< j len) (> level 0))
          (setq char (string-to-char (substring regexp j)))
          (cond
           ((char-equal char ?\\)
            (setq j (1+ j)
                  new-result
                  (cons (substring regexp (- j 2) j)
                        new-result)))
           ((char-equal char ?\])
            (setq level (1- level)))
           (t
            (setq new-result
                  (cons (char-to-string char) new-result))))
          (setq j (1+ j)))
        (and (not (= level 0)) (error "Malformed regexp."))
        (setq i j))
       ;; Zero or one repeat
       ((char-equal char ?\?)
        (setq new-result
              (cons ""
                    (if (listp new-result)
                        new-result
                      (list new-result)))))
       ;; Zero or more repeats
       ((char-equal char ?\*)
        (setq new-result (if (listp new-result) new-result (list new-result)))
        (let ((base new-result)
              (tmp-result new-result))
          (setq j 1)
          (or (and asterisk-max
                   (> asterisk-max 0))
              (error "Arg PLUS-MAX must be > 0 to use '+' in REGEXP."))
          (while (< j asterisk-max)
            (setq tmp-result (multiply-lists tmp-result base)
                  new-result (nconc new-result tmp-result)
                  j (1+ j)))
          (setq new-result
                (cons ""
                      (if (listp new-result)
                          new-result
                        (list new-result))))))
       ;; One or more repeats
       ((char-equal char ?\+)
        (setq new-result (if (listp new-result) new-result (list new-result)))
        (let ((base new-result)
              (tmp-result new-result))
          (setq tmp-result new-result
                j 1)
          (or (and plus-max
                   (> plus-max 0))
              (error "Arg PLUS-MAX must be > 0 to use '+' in REGEXP."))
          (while (< j plus-max)
            (setq tmp-result (multiply-lists tmp-result tmp-result)
                  new-result (nconc new-result tmp-result)
                  j (1+ j)))))
       ;; End of line
       ((char-equal char ?$)
        nil)
       ;; Character
       (t
        (setq result (multiply-lists result new-result)
              new-result (char-to-string char)))))
    (setq result (multiply-lists result new-result))
    (message nil)
    result))


;;; FILE NOT FOUND CODE 

(defun crypt-file-not-found-hook ()

  ;; Attempt at file-not-found support for encoded and encrypted files.

  ;; Loops through all elements of `crypt-encoding-alist' and the 
  ;; `crypt-encryption-type' element of `crypt-encryption-alist'. 
  ;; 
  ;; For each element of the alist it selects the 3rd element corresponding to 
  ;; file-name extensions. 
  ;; 
  ;; This file-name extension is in the form of a regular expression such as 
  ;; \\(\\.g?z\\)$
  ;; 
  ;; This is expanded into a list of strings using Lars Lindberg's 
  ;; `unregexp-quote' such as '(".z" ".gz") 
  ;; 
  ;; For each string in this new list of strings we append to the failed 
  ;; `buffer-file-name' and search for this new file name. 
  ;; 
  ;; If the augmented filename is found then the buffer for the old name is
  ;; killed and the routine `crypt-find-file-noselect' is called with the new
  ;; filename.
  
  (let ((alist
         (append crypt-encoding-alist
                 (list (assoc crypt-encryption-type crypt-encryption-alist))))
        (case-fold-search nil)
        file eltre)

    ;; Search through the file name extensions for a match.
    (while (and (not file)
                alist
                (setq eltre (nth 3 (car alist))))

      ;; Check to see if file exists sans extension. 
      (if (and (string-match eltre buffer-file-name)
               (file-readable-p
                (substring buffer-file-name 0 (match-beginning 1))))

          ;; It is a plain file.  Extract extension from file name.
          (setq file (substring buffer-file-name 0 (match-beginning 1)))

        ;; Expand the regular expression for the filename extension into a 
        ;; list of possible strings and append each to the failed filename
        ;; by looping over all variations that unregexp-quote returns.
        (let ((elt (unregexp-quote eltre)))
          (while (and elt
                      (not (file-readable-p
                            (setq file (concat buffer-file-name (car elt))))))
            (setq file nil
                  elt (cdr elt)))

          ;; Decrement, not needed for THEN form.
          (setq alist (cdr alist)))))
    
    ;; Did we find a match? 
    (if file
        (progn

          ;; Kill the old buffer for the wrong file.
          (kill-buffer buf)

          ;; Rename the file sought after.
          (setq filename file)

          ;; Read in the new file.
          (crypt-find-file-noselect filename)

          ;; Make sure the buffer-file-name of the buffer is set. Not needed?
          (set-buffer buf)
          (setq buffer-file-name file)

          ;; Return non-nil so `find-file-not-found-hooks' loop stops.
          t))))

;;; The following function `crypt-find-file-noselect' is based on
;;; `find-file-noselect' up to and including the insertion of the file
;;; contents.  This is done so that various file attributes are correctly
;;; handled for the new filename.  This is called by
;;; `crypt-file-not-found-hook'.

(defun crypt-find-file-noselect (filename)

  ;; Read file FILENAME into a buffer and return the buffer.  If a buffer
  ;; exists visiting FILENAME, return that one, but verify that the file has
  ;; not changed since visited or saved.  The buffer is not selected, just
  ;; returned to the caller.

  ;; We do not define any variables used since the idea is that we are trying 
  ;; to modify those of `find-file-noselect'

  (setq filename
        (abbreviate-file-name
         (expand-file-name filename)))
  (if (file-directory-p filename)
      (if find-file-run-dired
          (dired-noselect filename)
        (error "%s is a directory." filename))
    (setq buf (get-file-buffer filename)
          truename (abbreviate-file-name (file-truename filename))
          number (nthcdr 10 (file-attributes truename))
          ;; Find any buffer for a file which has same truename.
          same-truename
          (or buf ; Shortcut
              (let (found
                    (list (buffer-list)))
                (while (and (not found) list)
                  (save-excursion
                    (set-buffer (car list))
                    (if (and buffer-file-name
                             (string= buffer-file-truename truename))
                        (setq found (car list))))
                  (setq list (cdr list)))
                found))
          same-number
          (or buf ; Shortcut
              (and number
                   (let (found
                         (list (buffer-list)))
                     (while (and (not found) list)
                       (save-excursion
                         (set-buffer (car list))
                         (if (and (equal buffer-file-number number)
                                  ;; Verify this buffer's file number
                                  ;; still belongs to its file.
                                  (file-exists-p buffer-file-name)
                                  (equal (nthcdr 10 (file-attributes buffer-file-name)) number))
                             (setq found (car list))))
                       (setq list (cdr list)))
                     found)))
          error)
    ;; Let user know if there is a buffer with the same truename.
    (if (and (not buf) same-truename (not nowarn))
        (message "%s and %s are the same file (%s)"
                 filename (buffer-file-name same-truename)
                 truename)
      (if (and (not buf) same-number (not nowarn))
          (message "%s and %s are the same file"
                   filename (buffer-file-name same-number))))

    ;; Optionally also find that buffer.
    (if (or find-file-existing-other-name find-file-visit-truename)
        (setq buf (or same-truename same-number)))
    (if buf
        (or nowarn
            (verify-visited-file-modtime buf)
            (cond ((not (file-exists-p filename))
                   (error "File %s no longer exists!" filename))
                  ((yes-or-no-p
                    (format
                     (if (buffer-modified-p buf)
                         "File %s changed on disk.  Discard your edits? "
                       "File %s changed on disk.  Read the new version? ")
                     (file-name-nondirectory filename)))
                   (save-excursion
                     (set-buffer buf)
                     (revert-buffer t t)))))
      (save-excursion
        (setq buf (create-file-buffer filename))
        (set-buffer buf)
        (erase-buffer)
        (insert-file-contents filename t))))
  buf)

;;; Add ourselves to the `find-file-not-found-hooks' 
(add-hook 'find-file-not-found-hooks 'crypt-file-not-found-hook)


;;; Provide this package.
(provide 'crypt++-fnf)

;;; crypt++-fnf.el ends here
