;; mh-util.el
;; $Header: mh-utils.el,v 1.3 93/01/28 15:58:05 darrylo Exp $
;;
;;  Copyright (C) 1990 Free Software Foundation, Inc.
;;
;;     Author:  Alan Stebbens <aks@hub.ucsb.edu>
;;	Please send suggestions and corrections to the above address.
;;
;;; GNU Emacs is distributed in the hope that it will be useful, but
;;; without any warranty.  No author or distributor accepts
;;; responsibility to anyone for the consequences of using it or for
;;; whether it serves any particular purpose or works at all, unless he
;;; says so in writing.

;;; Everyone is granted permission to copy, modify and redistribute GNU
;;; Emacs, but only under the conditions described in the document "GNU
;;; Emacs copying permission notice".  An exact copy of the document is
;;; supposed to have been given to you along with GNU Emacs so that you
;;; can know how you may redistribute it all.  It should be in a file
;;; named COPYING.  Among other things, the copyright notice and this
;;; notice must be preserved on all copies.
;;
;; This file contains enhancements to the mh-e package, a GNU Emacs 
;; front end to the MH mail system, providing:
;;
;;  mh-invisible-headers-list  - a simple list of headers to ignore
;;  mh-set-invisible-headers   - build the regexp from headers list
;;  mh-toggle-headers          - show all/default headers
;;  mh-next-msg-same-kind      - next message with the same {sender,subject} 
;;  mh-prev-msg-same-kind      - prev message with the same {sender,subject}
;;  mh-scan-delete-by-subject  - delete by subject using scan listing 
;;  mh-scan-delete-by-body     - delete by body text using scan listing   
;;  mh-scan-delete-by-sender   - delete by sender using scan listing 
;;  mh-pick-delete-by-subject  - delete by subject using pick 
;;  mh-pick-delete-by-body     - delete by body text using pick 
;;  mh-pick-delete-by-sender   - delete by sender using pick 
;;  mh-do-pick-delete          - delete by matching pick pattern
;;
;; See the end of this file for the new key bindings.
;;
;; This file also fixes a bug in mh-next-pick-field, which refuses to allow
;; colons in the value of any component.
;;
;; Last Edited:
;; 
;; Wed Nov 18 22:55:12 1992 by Alan Stebbens (aks at anywhere.ucsb.edu)
;; 	 Cleaned up documentation.
;;
;; 	 Created mh-set-cur-scan-vars to facilitate debugging.
;;
;; 	 Fixed the mh-find-sender routine to use a substring of the
;; 	 sender if it is longer than the length of the sender field in
;; 	 the scan buffer.
;;
;; 	 Added some changes by John Dilley <jad@iag.hp.com>.
;; 	 Synopsis: Changed defconst to defvar so users can override them
;; 	 before loading this package (a defconst will override a
;; 	 previous setq).  Moved the assignment of string variables to
;; 	 separate setq statements.  This allows users of this package to
;; 	 change the value of some of the control variables and still
;; 	 have the following computed values correct after loading mh-util.
;; 
;; 
;; Thu Aug  8 13:58:47 1991 by Alan Stebbens (aks at anywhere.ucsb.edu)
;; 	 Fixed mh-scan-delete-by-*.
;; 
;; Sat Mar 23 16:28:10 1991 by Alan Stebbens (aks at anywhere.ucsb.edu)
;; 	 Fixed mh-scan-delete-by-kind to use ".*" before kind.
;; 
;; Thu Feb 28 18:25:10 1991 by Alan Stebbens (aks at anywhere.ucsb.edu)
;; 	 Added mh-scan-delete-by-XXXX funcs.
;; 
;; Wed Jun  6 14:02:36 1990 by Alan Stebbens (aks at somewhere.ucsb.edu)
;; 	 Boundary condition on backward subject searches fixed.
;; 
;; Mon Jun  4 21:16:26 1990 by Alan Stebbens (aks at somewhere.ucsb.edu)
;; 	 Added mh-{next,prev,find}-same-{kind,sender,subject},
;;	 and mh-match-msg.
;;	 Made mh-do-pick-delete call mh-next-msg at end.
;; 
;; Fri May 25 11:19:06 1990 by Alan Stebbens (aks at somewhere.ucsb.edu)
;; 	 Included mh-next-pick-field to fix bug (until fixed in
;; 	 mh-e.el).
;; 
;; Fri May 25 10:22:39 1990 by Alan Stebbens (aks at somewhere.ucsb.edu)
;; 	 Initial version.
;;

(require 'mh-e)				; be sure mh-e is loaded first


;; The following list may be predefined in your .emacs file.  If not, you get
;; these invisible headers by default.

(defvar mh-invisible-headers-list 
  '("Received"  	"Message-Id"	"Mail-From"
    "Return-Path"	"In-Reply-To"	"Resent-Message-Id"
    "Posted-Date"	"X-[^ \t]+"	"X400-[^ \t]+")
  "*List of message header names (which may contain regexp patterns) 
which will not be shown when displaying messages.  The headers in this
list will be used to construct the variable `mh-invisible-headers'.")

;; This function will cause the header list above to be used in creating
;; the invisible header pattern used by mh-e.

(defun mh-set-invisible-headers ()
  "This function causes the variable mh-invisible-headers to be rebuilt
according to the current list of headers in mh-invisible-headers-list."
  (interactive)
  (let ((hl mh-invisible-headers-list) pat hdr)
    (while hl
      (if pat (setq pat (concat pat "\\|")))
      (setq hdr (car hl))
      (setq hl (cdr hl))
      (setq pat (concat pat "^" hdr ":[ \t]*")))
    (setq mh-invisible-headers pat)))

(mh-set-invisible-headers)

;;
;; mh-toggle-headers
;;
;; Display or not, all visible headers

(defvar mh-visible-headers-obtained nil)
(defvar mh-orig-visible-headers nil)


(defun mh-toggle-headers (arg) "\
Set display of all message headers according to ARG: if nil, toggle the
current value; if 0, reset to default value (nil); if 1, show all
message headers."
  (interactive "p")
  (or mh-visible-headers-obtained
      (setq mh-orig-visible-headers mh-visible-headers
	    mh-visible-headers-obtained t))

  (setq mh-visible-headers
        (cond ((eq arg 4)                       ; verbose
               ".*")
              ((eq arg 0)                       ; original value
               mh-orig-visible-headers)
              ((eq arg 1)                       ; toggle original/verbose
               (if (string= mh-visible-headers ".*")
                   mh-orig-visible-headers
                 ".*"))))

  (save-excursion
    (if (get-buffer mh-show-buffer)
	(kill-buffer mh-show-buffer)))
  (mh-show (mh-get-msg-num t)))



(defvar mh-kind-search-default 'subject "\
*This variable should be set to either 'subject or 'sender, indicating the
default kind of search when used with \\[mh-next-msg-same-kind] or
\\[mh-prev-msg-same-kind].")

(defvar mh-delete-search-method 'scan "\
*This variable should be set to either 'scan or 'pick, indicating the default
method to use for the `mh-delete-by-...' functions.  The value is
toggled by \\[mh-toggle-delete-search-method].")

;; The following constants tailor the location in the folder scan listing of
;; the subject and body.  If you ever change the format of the scan listing
;; produced by mh-scan-folder, then you may have to alter these variables also.
;; They are defvar, not defconst, so users can override them before loading
;; this package.  A defconst will override a previous setq.
;; If you alter these variables interactively (say when you're debugging your
;; particular setup), then also do M-x mh-set-cur-scan-vars to compute the
;; dependent variables.

(defvar mh-cur-sender-scan-length nil
  "Length of the sender field in a scan listing. Set by mh-set-cur-scan-vars.")

(defvar mh-cur-sender-offset 12
  "*Offset in the current folder scan where the sender name begins.")

(defvar mh-cur-subject-offset 31
  "*Offset in the current folder scan where the subject begins.")

(defvar mh-cur-scan-sender-regexp nil
  "Regexp to match the sender portion of the current message.")

(defvar mh-cur-scan-subject-regexp nil
  "*Regexp which matches the subject of the current scan line.")

(defvar mh-good-sender-regexp nil
  "*Regexp used to match good message up to the sender portion.")

(defvar mh-good-subject-regexp nil
  "*Regexp matching a good message up to the subject portion.")

(defvar mh-cur-scan-body-regexp "<<[ \t]*\\(.*\\)"
  "*Regexp which matches the body included as part of the current scan line.")

(defvar mh-delete-body-prompt-length 50
  "*Default length of the initial substring prompt for \\[mh-delete-by-body]")

;;;;
;;  Set the above variables which were initialized to nil.  This is to allow
;;  users of this package to change the value of some of these variables and
;;  still have the following computed values correct after loading mh-util.
;;;;
(defun mh-set-cur-scan-vars ()
  "Function to set the mh-cur-scan-xxx variables; used during initialization
and interactively after changing any of the user variables from which the
former are computed."
  (interactive)
  (setq mh-cur-sender-scan-length (- mh-cur-subject-offset 
				     mh-cur-sender-offset)
	mh-cur-scan-sender-regexp
	(concat "^" (make-string mh-cur-sender-offset ?\. )
		"\\(" (make-string mh-cur-sender-scan-length ?\. )
		"\\)")
	mh-cur-scan-subject-regexp
	(concat "^" (make-string mh-cur-subject-offset ?\. )
		"\\(<?[^<\n]+\\)\\(<<\\|$\\)")
	mh-good-sender-regexp
	(concat mh-good-msg-regexp
		(make-string (- mh-cur-sender-offset (1+ mh-cmd-note)) ?\. ))
	mh-good-subject-regexp
	(concat mh-good-msg-regexp
		(make-string (- mh-cur-subject-offset (1+ mh-cmd-note)) ?\. )
		"\\(Re:[ \t]*\\|\\).*")
	))
(mh-set-cur-scan-vars)


;; mh-next-msg-same

(defun mh-next-msg-same-kind (arg) "\
Search forward in the current folder for another message with the same
subject, as the current message.  If prefix ARG is given, search by sender
name rather than subject."
  (interactive "P")
  (funcall (mh-same-kind-func arg) 'forward))

;; mh-prev-same-subject

(defun mh-prev-msg-same-kind (arg) "\
Search backward in the current folder for another message with the same
subject as the current message.  If prefix ARG is given, search by sender
name rather than subject."
  (interactive "P")
  (funcall (mh-same-kind-func arg) 'backward))

;; mh-same-kind-func

(defun mh-same-kind-func (arg) 
  (symbol-function 
   (intern (concat "mh-find-msg-same-" 
		   (if arg 
		       (if (eq mh-kind-search-default 'sender) 
			   "subject" "sender")
		     (if (eq mh-kind-search-default 'sender) 
			 "sender" "subject"))))))

(defun mh-toggle-kind-search () "\
Toggle the kind of message search between 'subject and 'sender."
  (interactive)
  (setq mh-kind-search-default 
	(if (eq mh-kind-search-default 'sender) 'subject 'sender))
  (message "Next default searches will be %s kind" (symbol-name mh-kind-search-default)))

;; mh-find-same-subject

(defun mh-find-msg-same-subject (direction) "\
Search the current scan buffer according to DIRECTION ('forward or 'backward)
for another message with the same subject as the current message."
  (let* ((subject (mh-current-subject))
	 (regexp (concat mh-good-subject-regexp 
			 (regexp-quote subject))))
    (mh-match-msg direction regexp "subject" subject)))

;; mh-find-msg-same-sender

(defun mh-find-msg-same-sender (direction) "\
Search the current folder's scan buffer given DIRECTION for another message
with the same sender."
  (let ((sender (mh-current-sender))
	regexp)
    (if (string-match "\\`\\(.+[^ \t]\\)[ \t]+\\'" sender)
	(setq sender (substring sender 0 (match-end 1))))
    (if (>= (length sender) mh-cur-sender-scan-length)
	(setq sender (substring sender 0 mh-cur-sender-scan-length)))
    (setq regexp (concat mh-good-sender-regexp ".*" (regexp-quote sender)))
    (mh-match-msg direction regexp "sender" sender)))

;; mh-match-msg

(defun mh-match-msg (direction regexp component value) "\
Find the next message given DIRECTION matching the given REGEXP.
Third and fourth args are COMPONENT and VALUE, for error messages
on search failure."
  (let* ((dir (symbol-name direction))
	 (search (intern (concat "re-search-" dir)))
	 (msg (save-excursion
		(beginning-of-line (if (eq direction 'forward) 2 1))
		(if (funcall search regexp nil t)
		    (mh-get-msg-num nil)))))
    (if (and msg (mh-goto-msg msg))
	(setq mh-next-direction direction)
      (message "No more messages %s with %s: \"%s\"" dir component value)
      (ding))))

;; There are two categories of deletion operations: 
;;   use the scan listing
;;   use the `pick' command (which works regardless of the validity of
;;   the scan listing).
;;
;; The current delete mode can be set by using M-x mh-toggle-delete-mode.
;;
;; Given the category, there are three kinds of deletion operations:
;;
;;  delete by sender
;;  delete by subject
;;  delete by body
;;
;; Delete by sender and subject can be accomplished with either the scan 
;; listing, or by using mh-do-pick-delete.  Delete by body is best done by
;; using mh-do-pick-delete.

;; mh-toggle-delete-search-method

(defun mh-toggle-delete-search-method (arg) "\
Toggle the search method for message deletion according to ARG: if nil, 
toggle the current value; if t, use the 'scan methods, otherwise, use
the 'pick methods.  This function sets the value of the variable 
\"mh-delete-search-method\"."
  (interactive "P")
  (setq mh-delete-search-method
	(cond ((null arg)
	       (if (eq 'scan mh-delete-search-method)
		   'pick
		 'scan))
	      ((eq t arg)
	       'scan)
	      (t 'pick)))
  (message "Deletion searches will use the \"%s\" method." 
	   (symbol-name mh-delete-search-method)))

;; mh-delete-search-func

;; Return a function which performs the current deletion method for the given
;; ARG, which *must* be one of the strings: "subject", "sender", or "body".

(defun mh-delete-search-func (arg)
  (symbol-function
   (intern (concat "mh-" 
		   (if (eq mh-delete-search-method 'scan)
		       "scan-delete-by-"
		     "pick-delete-by-")
		   arg))))

;; mh-delete-by-arg-list

;; Return an argument list suitable to either the `scan' or `pick'
;; -delete-by-METHOD functions.

(defun mh-delete-by-arg-list (kind default)
  "Return a list representing the arguments to be passed to one of the 
delete/find-by-{subject,sender,body} functions."
  (list current-prefix-arg
	(read-string 
	 (if (eq 'scan mh-delete-search-method)
	     (format "%s scan delete by %s: "
		     (capitalize (symbol-name (if current-prefix-arg 
						  (if (eq 'forward mh-next-direction)
						      'backward 'forward)
						mh-next-direction)))
		     kind)
	   (format "Pick delete by %s: " kind))
	 default)))

;; mh-delete-by-subject
;; mh-delete-by-sender
;; mh-delete-by-body

;; These functions delete the subsequent messages, using either the "pick"
;; method or the "scan" method, depending upon the variable
;; mh-delete-search-method.

(defun mh-delete-by-subject (arg subject) "\
Delete the messages with the same subject as SUBJECT, with the default being
the subject of the current message.  The search method will be either to use
the current \"scan\" folder listing, or to invoke \"pick\" with the appropriate
arguments.  If using a `scan' delete, the prefix ARG toggles the most recent 
search direction.  See the variable \"mh-delete-search-method\" for the current 
default search method; see the functions \"mh-scan-delete-by-subject\" 
and \"mh-pick-delete-by-subject\"."
  (interactive (mh-delete-by-arg-list "subject" (mh-current-subject)))
  (funcall (mh-delete-search-func "subject") arg subject))

(defun mh-delete-by-sender (arg sender) "\
Delete the messages with the same sender as SENDER, with the default being
the sender of the current message.   The search method will be either to use
the current \"scan\" folder listing, or to invoke \"pick\" with the appropriate
arguments.  If using a `scan' delete, the prefix ARG toggles the most recent
search direction.  See the variable \"mh-delete-search-method\" for the current
default search method; see the functions \"mh-scan-delete-by-sender\" and
\"mh-pick-delete-by-sender\"."
  (interactive (mh-delete-by-arg-list "sender" (mh-current-sender)))
  (funcall (mh-delete-search-func "sender") arg sender))

(defun mh-delete-by-body (arg body) "\
Delete the messages with the same body as BODY, with the default being
the body of the current message.   The search method will be either to use
the current \"scan\" folder listing, or to invoke \"pick\" with the appropriate
arguments.  If using a `scan' delete, the prefix ARG toggles the most recent
search direction.  See the variable \"mh-delete-search-method\" for the current
default search method; see the functions \"mh-scan-delete-by-body\" and
\"mh-pick-delete-by-body\"."
  (interactive (mh-delete-by-arg-list "body" (mh-current-body)))
  (funcall (mh-delete-search-func "body") arg body))

;;
;; mh-scan-delete-by-subject

(defun mh-scan-delete-by-subject (arg subject) "\ 
Search the current folder for all messages of the same subject as the
current message and mark them for deletion."
  (mh-scan-delete-by-kind subject 
			  mh-good-subject-regexp 
			  mh-cur-scan-subject-regexp
			  arg))

(defun mh-scan-delete-by-sender (arg sender) "\
Search the current folder for all messages of the same sender as the
current message and mark them for deletion."
  (mh-scan-delete-by-kind sender
			  mh-good-sender-regexp
			  mh-cur-scan-sender-regexp
			  arg))

(defun mh-scan-delete-by-body (arg body) "\
Search the current folder for all messages with the same body pattern
as the current message and mark them for deletion."
  (mh-scan-delete-by-kind body
			  mh-good-msg-regexp
			  mh-cur-scan-body-regexp
			  arg))

(defun mh-scan-delete-by-kind (kind search-re match-re arg)
  "Search the current scan listing for all messages of the same KIND using
the SEARCH-RE (a regular expression) to search for good messages and 
MATCH-RE to match the KIND."
  (let* ((deleted 0)
	(dir (if arg (if (eq 'forward mh-next-direction)
			 'backward 'forward)
	       mh-next-direction))
	(search (symbol-function (intern (concat "search-" (symbol-name dir)))))
	(move (symbol-function (intern (concat (if (eq dir 'backward) "previous" "forward") "-line"))))
	msgnum)
    (save-excursion
      (message "Searching...")
      (while (funcall search kind nil t)
        (if (eq dir 'backward) 
            (forward-line 1))
	(beginning-of-line 1)
	(if (and (looking-at search-re)
		 (looking-at match-re)
		 (string-match (concat ".*" (regexp-quote kind))
			       (buffer-substring (match-beginning 1) 
						 (match-end 1))))
	    (progn
	      (mh-delete-a-msg (setq msgnum (mh-get-msg-num nil)))
	      (message "Deleted message %d..." msgnum)
	      (setq deleted (1+ deleted))))
	(funcall move 1)))
    (if (eq dir 'forward) 
	(mh-next-undeleted-msg)
      (mh-previous-undeleted-msg))
    (message "Deleted %d messages" deleted)))


;; mh-pick-delete-by-subject
;; Delete the current msg and any others with the same subject.

(defun mh-pick-delete-by-subject (arg subject) "\
Search the current folder for messages with the given SUBJECT and delete
them."
  (interactive (list current-prefix-arg
		     (read-string "Delete by subject: " (mh-current-subject))))
  (let ((folder mh-current-folder))
    (set-buffer (get-buffer-create " delete-pattern"))
    (mh-make-pick-template)
    (mh-insert-fields "Subject:" subject)
    (setq mh-searching-folder folder)
    (mh-do-pick-delete)))

;; mh-pick-delete-by-sender
;; Delete the current msg and any others with the same sender.

(defun mh-pick-delete-by-sender (arg sender) "\
Search the current folder for messages with the given SENDER and delete them."
  (interactive (list current-prefix-arg
		     (read-string "Delete by sender: " (mh-current-sender))))
  (let ((folder mh-current-folder))
    (set-buffer (get-buffer-create " delete-pattern"))
    (mh-make-pick-template)
    (mh-insert-fields "From:" sender)
    (setq mh-searching-folder folder)
    (mh-do-pick-delete)))

;; mh-pick-delete-by-body
;; Delete any messages with the same body text as the current message.

(defun mh-pick-delete-by-body (arg body) "\
Search the current folder for messages with the given BODY and delete them.
Prompt for BODY if not supplied."
  (interactive (list current-prefix-arg 
		     (read-string "Delete by text: " (mh-current-body))))
  (let ((folder mh-current-folder))
    (set-buffer (get-buffer-create " delete-pattern"))
    (mh-make-pick-template)
    (mh-goto-header-end 1)
    (insert body)
    (setq mh-searching-folder folder)
    (mh-do-pick-delete)))

;; mh-current-sender

(defun mh-current-sender ()
  "Get the current sender, either from the current message, if it is 
being shown, or from the folder scan listing."
  (let ((sender  (if (and mh-showing (eq mh-delete-search-method 'pick))
		     (save-window-excursion
		       (set-buffer mh-show-buffer)
		       (save-excursion
			 (or (mh-get-field "From:")
			     (mh-get-field "Sender:")
			     (mh-get-field "Reply-To:"))))
		   (save-excursion
		     (beginning-of-line)
		     (if (looking-at mh-cur-scan-sender-regexp)
			 (buffer-substring (match-beginning 1) (match-end 1)))))))
    (if (string-match "[ \t]+$" sender)
	(setq sender (substring sender 0 (match-beginning 0))))
    sender))

;; mh-current-subject

(defun mh-current-subject ()
  "Get the current subject, either from the folder scan listing, or from
the current message, if it is being shown."
  (let ((subject (if (and mh-showing (eq mh-delete-search-method 'pick))
		     (save-window-excursion
		       (set-buffer mh-show-buffer)
		       (save-excursion
			 (mh-get-field "Subject:")))
                   (save-excursion
		     (beginning-of-line)
		     (if (looking-at mh-cur-scan-subject-regexp)
			 (buffer-substring (match-beginning 1) (match-end 1)))))))
    (if (string-match "^[ \t]*[Rr]e:[ \t]*" subject)
	(setq subject (substring subject (match-end 0))))
    (if (string-match "[ \t]+$" subject)
	(setq subject (substring subject 0 (match-beginning 0))))
    subject))

;; mh-current-body

(defun mh-current-body ()
  "Get the current message body, either from the folder scan listing, or
from the current message, if it is being shown."
  (let ((str (if mh-showing 
		 (save-window-excursion
		   (set-buffer mh-show-buffer)
		   (save-excursion
		     (if (and (mh-goto-header-end 1)
			      (re-search-forward "[^ \t\n]" nil t))
			 (let (beg end)
			   (setq beg (1- (point)))
			   (end-of-line)
			   (buffer-substring beg (point))))))
	       (save-excursion 
		 (let (end)
		   (end-of-line)
		   (setq end (point))
		   (beginning-of-line)
		   (if (re-search-forward mh-cur-scan-body-regexp end t)
		       (buffer-substring (match-beginning 1) (match-end 1))))))))
    (if (and str (> (length str) mh-delete-body-prompt-length))
	(substring str 0 mh-delete-body-prompt-length)
      str)))
    
;; mh-do-pick-delete
;;
;; Take the current pick-pattern buffer and use it to search and match messages
;; to delete.
;; (copied from mh-do-pick-search, but we use a different sequence name: 'delete)

(defun mh-do-pick-delete ()
  "Find messages that match the qualifications in the current pattern buffer.
Messages are searched for in the folder named in mh-searching-folder.
Delete messages found."
  (interactive)
  (let ((pattern-buffer (buffer-name))
	(searching-buffer mh-searching-folder)
	range pattern new-buffer msgs)
    (save-excursion
      (cond ((get-buffer searching-buffer)
	     (set-buffer searching-buffer)
	     (setq range (format "%d-%d" mh-first-msg-num mh-last-msg-num)))
	    (t
	     (mh-make-folder searching-buffer)
	     (setq range "all")
	     (setq new-buffer t))))
    (message "Searching %s..." searching-buffer)
    (goto-char (point-min))
    (while (setq pattern (mh-next-pick-field pattern-buffer))
      (setq msgs (mh-seq-from-command searching-buffer
				      'delete
				      (nconc (cons "pick" pattern)
					     (list searching-buffer
						   range
						   "-sequence" "delete"
						   "-list"))))
      (setq range "delete"))
    (message "Searching %s...done" searching-buffer)
    (if new-buffer
	(mh-scan-folder searching-buffer msgs)
	(switch-to-buffer searching-buffer))
    (delete-other-windows)
    (message "Deleting %d messages..." (length msgs))
    (mh-delete-msg-no-motion 'delete)
    (message "Deleted %d messages." (length msgs))
    (mh-next-msg)))

;;
;; Fixes the bug in the standard mh-e.el (modified the regexp).
;;

(defun mh-next-pick-field (buffer)
  ;; Return the next piece of a pick argument that can be extracted from the
  ;; BUFFER.  Returns nil if no pieces remain.
  (set-buffer buffer)
  (let ((case-fold-search t))
    (cond ((eobp)
	   nil)
	  ((re-search-forward "^\\([a-z][^:\n \t]*\\):[ \t]*\\([a-z0-9].*\\)$" nil t)
	   (let* ((component
		   (format "-%s"
			   (downcase (buffer-substring (match-beginning 1)
						       (match-end 1)))))
		  (pat (buffer-substring (match-beginning 2) (match-end 2))))
	       (forward-line 1)
	       (list component pat)))
	  ((re-search-forward "^-*$" nil t)
	   (forward-char 1)
	   (let ((body (buffer-substring (point) (point-max))))
	     (if (and (> (length body) 0) (not (equal body "\n")))
		 (list "-search" body)
		 nil)))
	  (t
	   nil))))

;; Make new commands available by keystroke

(define-key mh-folder-mode-map "D"         'mh-delete-by-subject)
(define-key mh-folder-mode-map "\M-d"      'mh-delete-by-sender)
(define-key mh-folder-mode-map "\M-\C-d"   'mh-delete-by-body)
(define-key mh-folder-mode-map "h"	   'mh-toggle-headers)
(define-key mh-folder-mode-map "\M-t"	   'mh-toggle-headers)
(define-key mh-folder-mode-map "N"	   'mh-next-msg-same-kind)
(define-key mh-folder-mode-map "P"	   'mh-prev-msg-same-kind)
(define-key mh-folder-mode-map "T"	   'mh-toggle-kind-search)
(define-key mh-folder-mode-map "\M-\C-t"   'mh-toggle-delete-search-method)
(define-key mh-pick-mode-map   "\C-c\C-d"  'mh-do-pick-delete)

;; Announce that we're loaded

(provide 'mh-utils)
